package org.LiveGraph;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JOptionPane;

import org.LiveGraph.dataCache.DataCache;
import org.LiveGraph.dataCache.UpdateInvoker;
import org.LiveGraph.gui.DataFileSettingsWindow;
import org.LiveGraph.gui.GraphSettingsWindow;
import org.LiveGraph.gui.MainWindow;
import org.LiveGraph.gui.PlotWindow;
import org.LiveGraph.gui.SeriesSettingsWindow;
import org.LiveGraph.plot.GraphExporter;
import org.LiveGraph.plot.Plotter;
import org.LiveGraph.settings.DataFileSettings;
import org.LiveGraph.settings.DataSeriesSettings;
import org.LiveGraph.settings.ErrorWhileSettingHasChangedProcessingException;
import org.LiveGraph.settings.GraphSettings;

import com.softnetConsult.utils.exceptions.ThrowableTools;


/**
 * This is the main executable class of the LiveGraph plotter application.
 * An instance of this class represents the application itself. The tasks of this
 * class is to interpret the command line parameters, to set-up and to start-up
 * the GUI and the back-end of the application, and to provide some
 * functions which are used by different modules of the application to communicate
 * with each other and to access global data, such as settings.
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: LiveGraph.java</p>
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p> 
 *  
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class LiveGraph {

/**
 * LiveGraph software version.
 */
public static final String version = "1.1.1";

// Static stuff {

/**
 * Singleton application instance.
 */
private static LiveGraph app = null;

/**
 * Singleton application access method.
 * 
 * @return The singleton application object.
 */
public static LiveGraph application() {
	if (null == LiveGraph.app)
		LiveGraph.app = new LiveGraph();
	return LiveGraph.app;
}

/**
 * Program entry point.
 * Creates an application instance and calls the {@link #exec(String[])} method.
 * 
 * @param args Command line parameters.
 */
public static void main(String [] args) {	
	LiveGraph.application().exec(args);
}

// } end of static stuff.


/**
 * Application's data update invoker.
 */
private UpdateInvoker updateInvoker = null;


/**
 * Main GUI window of the application. 
 */
private MainWindow mainWindow = null;

/**
 * GUI window for data file settings.
 */
private DataFileSettingsWindow dataFileSettingsWindow = null;

/**
 * GUI window for graph settings.
 */
private GraphSettingsWindow graphSettingsWindow = null;

/**
 * GUI window for data series settings.
 */
private SeriesSettingsWindow seriesSettingsWindow = null;

/**
 * GUI window for the actual graph plot.
 */
private PlotWindow plotWindow = null;


/**
 * Holds the data file settings for the application.
 */
private DataFileSettings dataFileSettings = null;

/**
 * Holds the graph settings for the application.
 */
private GraphSettings graphSettings = null;

/**
 * Holds the data series settings for the application.
 */
private DataSeriesSettings seriesSettings = null;

/**
 * Holds the graph exporter.
 */
private GraphExporter graphExporter = null;


/**
 * Main program method.
 * It parses the command line parameters, sets up the GUI and the back-end components
 * of the application and configures the their communication. It then loads the default
 * settings and passes the execution control to the main Swing GUI loop. 
 * 
 * @param args Command line arguments.
 */
public void exec(String [] args) {
	
	// Setup exception handling:
	installUncaughtExceptionHandler();
	
	// Buffer for names of the settings files to load automatically:
	Map<String, String> startupFiles = new HashMap<String, String>();
	
	// Parse the command line arguments and combine the results with default settings:
	String argsParseErrorMsg = getInitialSettingsFiles(args, startupFiles);
	
	// Create settings holder objects:
	dataFileSettings = new DataFileSettings();
	graphSettings = new GraphSettings();
	seriesSettings = new DataSeriesSettings();
	
	// Set-up communication between different settings objects:
	graphSettings.addObserver(seriesSettings);
	
	// Create the data cache and the data update invoker,
	// set-up their communication with other objects,
	// and create the data update invocation thread:
	DataCache dataCache = new DataCache();
	updateInvoker = new UpdateInvoker();
	updateInvoker.setDataCache(dataCache);
	dataFileSettings.addObserver(updateInvoker);
	Thread fileUpdateInvokerThread = new Thread(updateInvoker, "Update invoker thread");
	
	// Create the graph plotter and set-up its communication with other objects:
	Plotter mainPlotter = new Plotter(dataCache);
	dataCache.addObserver(mainPlotter);
	graphSettings.addObserver(mainPlotter);
	seriesSettings.addObserver(mainPlotter);
	
	// Create the graphExporter:
	this.graphExporter = new GraphExporter(mainPlotter);
	
	// Create the main application window:
	mainWindow = new MainWindow();
	mainWindow.setVisible(true);
	
	// Create the data file settings window and set-up its communication with other objects:
	dataFileSettingsWindow = new DataFileSettingsWindow();
	setDisplayDataFileSettingsWindow(true);
	updateInvoker.addObserver(dataFileSettingsWindow);
	dataCache.addObserver(dataFileSettingsWindow);
	dataFileSettings.addObserver(dataFileSettingsWindow);
	
	// Create the graph settings window and set-up its communication with other objects:
	graphSettingsWindow = new GraphSettingsWindow();
	setDisplayGraphSettingsWindow(true);
	dataCache.addObserver(graphSettingsWindow);
	graphSettings.addObserver(graphSettingsWindow);
	
	// Create the data series settings window and set-up its communication with other objects:
	seriesSettingsWindow = new SeriesSettingsWindow();
	setDisplaySeriesSettingsWindow(true);
	dataCache.addObserver(seriesSettingsWindow);
	seriesSettings.addObserver(seriesSettingsWindow);
	graphSettings.addObserver(seriesSettingsWindow);
	
	// Create the graph plot window and set-up its communication with other objects:
	plotWindow = new PlotWindow(mainPlotter);
	setDisplayPlotWindow(true);
	plotWindow.addSeriesHighlightListeners(seriesSettingsWindow);
	dataCache.addObserver(plotWindow);
	seriesSettings.addObserver(plotWindow);
	graphSettings.addObserver(plotWindow);
	
	// Start the data update invocation thread:
	fileUpdateInvokerThread.start();
	
	// Display any possible error messages about the command line arguments:
	if (null != argsParseErrorMsg)
		logErrorLn(argsParseErrorMsg);
	
	// Load default or command-line specified graph settings:
	if (startupFiles.containsKey("GraphSettings")) {
		String fn = startupFiles.get("GraphSettings");
		try {
			if (graphSettings.load(fn))
				logSuccessLn("Initial graph settings loaded from \"" + fn + "\".");
			else
				logErrorLn("Error while loading graph settings from \"" + fn + "\".");
		} catch (ErrorWhileSettingHasChangedProcessingException e) {
			logErrorLn("There was a problem while loading initial graph settings: \n"
					 + "    " + (null != e.getCause() ? e.getCause().getMessage() : e.getMessage()) + ".");			
		}
	}
	
	// Load default or command-line specified data series settings:
	if (startupFiles.containsKey("DataSeriesSettings")) {
		String fn = startupFiles.get("DataSeriesSettings");
		try {
			if (seriesSettings.load(fn))
				logSuccessLn("Initial data series settings loaded from \"" + fn + "\".");
			else
				logErrorLn("Error while loading data series settings from \"" + fn + "\".");
		} catch (ErrorWhileSettingHasChangedProcessingException e) {
			logErrorLn("There was a problem while loading initial data series settings: \n"
					 + "    " + (null != e.getCause() ? e.getCause().getMessage() : e.getMessage()) + ".");
		}
	}
	
	// Load default or command-line specified data file settings:
	if (startupFiles.containsKey("DataFileSettings")) {
		String fn = startupFiles.get("DataFileSettings");
		try {
			if (dataFileSettings.load(fn))
				logSuccessLn("Initial data file settings loaded from \"" + fn + "\".");
			else
				logErrorLn("Error while loading data file settings from \"" + fn + "\".");
		} catch (ErrorWhileSettingHasChangedProcessingException e) {
			logErrorLn("There was a problem while loading initial data file settings: \n"
					 + "    " + (null != e.getCause() ? e.getCause().getMessage() : e.getMessage()) + ".");
			
		}
	}
	
	if (!runsCorrectJavaVersion()) {
		JOptionPane.showMessageDialog(null, "The Java runtime environment you are using may not "
										  + "support all program features.\n\n"
										  + "LiveGraph is targeted for Java version 1.6 or later, "
										  + "however, it may run on earlier Java versions with a "
										  + "reduced feature set.\nNote that various error messages "
										  + "may be displayed when accessing the unsupported features.\n\n"
										  + "Your current Java version is " + getJavaSpecificationVersion(),
									  "Incompatible Java version", JOptionPane.WARNING_MESSAGE);
	}
	
} // public void exec(String [] args)

/**
 * Determines the current Java specification version. 
 * @return The current Java specification version or {@code "unknown"} if it could not be obtained.
 */
public String getJavaSpecificationVersion() {
	try {
		return System.getProperty("java.specification.version");
	} catch (Throwable e) {
		return "unknown";
	}
}

/**
 * Determines whether the currect Java version is appropriate. This is done based on the system
 * property {@code java.specification.version}. Java version {@code 1.6} or higher is considered ok.
 * @return Whether the currect Java version is appropriate.
 */
public boolean runsCorrectJavaVersion() {
	
	String specVer = getJavaSpecificationVersion();	
		
	if (specVer.equalsIgnoreCase("unknown"))
		return false;
	
	int p = specVer.indexOf(".");
	if (0 > p)
		return false;
	
	int mainVer = Integer.parseInt(specVer.substring(0, p));
	if (1 > mainVer)
		return false;
	if (1 < mainVer)
		return true;
	
	if (specVer.length() - 1 <= p)
		return false;
	
	int subVer =  Integer.parseInt(specVer.substring(p + 1, p + 2));
	if (6 > subVer)
		return false;
	
	return true;
}

/**
 * Parses the command line arguments for file names for initial settings and combines the results
 * with default settings file names.
 * 
 * @param args Command line arguments.
 * @param startupFiles A table to hold the results.
 * @return {@code null} if no error occured or the error message if there was an error (e.g. incorrect
 * command line arguments).
 */
private String getInitialSettingsFiles(String [] args, Map<String, String> startupFiles) {

	String errMsg = getInitialSettingsFromCommandLine(args, startupFiles);
	setDefaultInitialSettingsFiles(startupFiles);
	return errMsg;
} // getInitialSettingsFiles

/**
 * Parses the command line arguments for file names for initial settings.
 * 
 * @param args Command line arguments.
 * @param startupFiles A table to hold the results.
 * @return {@code null} if no error occured or the error message if there was an error (e.g. incorrect
 * command line arguments).
 */
private String getInitialSettingsFromCommandLine(String [] args, Map<String, String> startupFiles) {
	
	if (0 != args.length && 2 != args.length && 4 != args.length && 6 != args.length) {
		return "Command line parameters are invalid and were ignored. \n"
			 + "    Legal command line arguments are as follows: \n"
			 + "    > java edu.monash.LiveGraph.LiveGraph [-dfs \"{data file settings file}\"] \n"
			 + "                                          [-gs \"{graph settings file}\"] \n"
			 + "                                          [-dss \"{data series settings file}\"] \n"
			 + "    This means the program expects either 0, 2, 4 or 6 command line arguments. \n"
			 + "    However, " + args.length + " argument" + (1==args.length?" was":"s were") + " passed.";
	}
	
	String errMsg = "";
	String s = null;
	
	if (2 <= args.length) {		
		s = tryParseCommandLineArgument(args[0], args[1], startupFiles);
		if (null != s && 0 < errMsg.length())
			s += "\n";
		if (null != s)
			errMsg += s;
	}
	
	if (4 <= args.length) {
		s = tryParseCommandLineArgument(args[2], args[3], startupFiles);
		if (null != s && 0 < errMsg.length())
			s += "\n";
		if (null != s)
			errMsg += s;
	}
	
	if (6 <= args.length) {
		s = tryParseCommandLineArgument(args[4], args[5], startupFiles);
		if (null != s && 0 < errMsg.length())
			s += "\n";
		if (null != s)
			errMsg += s;
	}
	
	if (0 == errMsg.length())
		return null;
	
	return errMsg;
} // getInitialSettingsFromCommandLine

/**
 * Processes two consecutive command line parameters.
 * 
 * @param flag Flag specifier command line parameter.
 * @param filename Filename specifier command line parameter.
 * @param startupFiles A table to hold the results.
 * @return {@code null} if no error occured or the error message if there was an error (e.g. incorrect
 * command line arguments).
 */
private String tryParseCommandLineArgument(String flag, String filename, Map<String, String> startupFiles) {
	
	if (flag.equalsIgnoreCase("-dfs")) {
		File f = new File(filename);
		if (!f.exists())
			return "Could not load the data file settings from the file specified on the command line "
			     + "because the file does not exist. \n"
			     + "    (" + filename + ")";
		startupFiles.put("DataFileSettings", f.getAbsolutePath());
		return null;
	}
	
	if (flag.equalsIgnoreCase("-gs")) {
		File f = new File(filename);
		if (!f.exists())
			return "Could not load the graph settings from the file specified on the command line "
			     + "because the file does not exist. \n"
			     + "    (" + filename + ")";
		startupFiles.put("GraphSettings", f.getAbsolutePath());
		return null;
	}
	
	if (flag.equalsIgnoreCase("-dss")) {
		File f = new File(filename);
		if (!f.exists())
			return "Could not load the graph settings from the file specified on the command line "
			     + "because the file does not exist. \n"
			     + "    (" + filename + ")";
		startupFiles.put("DataSeriesSettings", f.getAbsolutePath());
		return null;
	}
	
	return "Invalid command line flag \"" + flag + "\" will be ignored. \n"
		 + "    The subsequent command line argument \"" + filename + "\" will also be ignored.";
} // tryParseCommandLineArgument


/**
 * If some settings were not specified to be loaded from a file via the command line,
 * this method is used to specify default files for loading the settings, provided the
 * defualt files exist.
 * 
 * @param startupFiles A table to hold the results.
 */
private void setDefaultInitialSettingsFiles(Map<String, String> startupFiles) {
	
	if (! startupFiles.containsKey("DataFileSettings")) {
		File f = new File("session" + DataFileSettings.preferredFileExtension); 
		if (f.exists())
			startupFiles.put("DataFileSettings", f.getAbsolutePath());
	}
	
	if (! startupFiles.containsKey("GraphSettings")) {
		File f = new File("session" + GraphSettings.preferredFileExtension); 
		if (f.exists())
			startupFiles.put("GraphSettings", f.getAbsolutePath());
	}
	
	if (! startupFiles.containsKey("DataSeriesSettings")) {
		File f = new File("session" + DataSeriesSettings.preferredFileExtension); 
		if (f.exists())
			startupFiles.put("DataSeriesSettings", f.getAbsolutePath());
	}
} // setDefaultInitialSettingsFiles

/**
 * Prints an info message to the main window message area.
 * 
 * @param o The message.
 */
public void logInfoLn(Object o) {
	mainWindow.logInfoLn(o.toString());
}

/**
 * Prints an error message to the main window message area.
 * 
 * @param o The message.
 */
public void logErrorLn(Object o) {
	mainWindow.logErrorLn(o.toString());
}

/**
 * Prints an success message to the main window message area.
 * 
 * @param o The message.
 */
public void logSuccessLn(Object o) {
	mainWindow.logSuccessLn(o.toString());
}

/**
 * Displays or hides the data file settings window.
 * 
 * @param state Whether to display ({@code true}) or to hide ({@code false}).
 */
public void setDisplayDataFileSettingsWindow(boolean state) {	
	dataFileSettingsWindow.setVisible(state);
	mainWindow.fileSettingsDisplayStateChanged(state);
}

/**
 * Displays or hides the graph settings window.
 * 
 * @param state Whether to display ({@code true}) or to hide ({@code false}).
 */
public void setDisplayGraphSettingsWindow(boolean state) {	
	graphSettingsWindow.setVisible(state);
	mainWindow.graphSettingsDisplayStateChanged(state);
}

/**
 * Displays or hides the data series settings window.
 * 
 * @param state Whether to display ({@code true}) or to hide ({@code false}).
 */
public void setDisplaySeriesSettingsWindow(boolean state) {	
	seriesSettingsWindow.setVisible(state);
	mainWindow.seriesSettingsDisplayStateChanged(state);
}

/**
 * Displays or hides plot window.
 * 
 * @param state Whether to display ({@code true}) or to hide ({@code false}).
 */
public void setDisplayPlotWindow(boolean state) {	
	plotWindow.setVisible(state);
	mainWindow.plotDisplayStateChanged(state);
}

/**
 * This method is called by the main window when it is closed. This method
 * initiates the disposing of all windows and the data update invocation
 * thread in order to correctly close the application and save all settings
 * to default files. 
 */
public void disposeGUIAndExit() {
	mainWindow.dispose();
	dataFileSettingsWindow.dispose();
	graphSettingsWindow.dispose();
	seriesSettingsWindow.dispose();
	plotWindow.dispose();
	graphExporter.disposeInternalGUI();
	updateInvoker.setMustQuit(true);
	
	dataFileSettings.save("session" + DataFileSettings.preferredFileExtension);
	graphSettings.save("session" + GraphSettings.preferredFileExtension);
	seriesSettings.save("session" + DataSeriesSettings.preferredFileExtension);
}

/**
 * Gets the application's global data file settings.
 * 
 * @return Global data file settings.
 */
public DataFileSettings getDataFileSettings() {
	return dataFileSettings;
}

/**
 * Gets the application's global graph settings.
 * 
 * @return Global graph settings.
 */
public GraphSettings getGraphSettings() {
	return graphSettings;
}


/**
 * Gets the application's global data series settings.
 * 
 * @return Global data series settings.
 */
public DataSeriesSettings getDataSeriesSettings() {
	return seriesSettings;
}

public GraphExporter getGraphExporter() {
	return graphExporter;
}

/**
 * Causes the next data update.
 */
public void initiateDataUpdate() {
	updateInvoker.update();
}

private void installUncaughtExceptionHandler() {
	
	try {
		UncaughtExceptionHandler handler = new UncaughtExceptionHandler();
		Thread.setDefaultUncaughtExceptionHandler(handler);
		
	} catch (SecurityException e) {
		e.printStackTrace();
	}
}

private class UncaughtExceptionHandler implements Thread.UncaughtExceptionHandler {
	public void uncaughtException(Thread t, Throwable e) {
		try {
			String err = ThrowableTools.stackTraceToString(e);
			if (null == mainWindow) {
				String h = "Error in thread \"" + t.getName() + "\"";
				JOptionPane.showMessageDialog(null, err, h, JOptionPane.ERROR_MESSAGE);
			} else {
				logErrorLn(err);
			}
		} finally {
			e.printStackTrace();
		}
		//throw new Error(e);
    }
} // private class UncaughtExceptionHandler

} // public class LiveGraph
