package org.LiveGraph.dataFile.write;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.util.Calendar;
import java.util.GregorianCalendar;



/**
 * This class provides static convenience methods for creating dedicated data stream writers.
 * Given just a directory on the local hard drive, this class can automatically
 * choose a descriptive and unique name for a data file and return an appropriate
 * {@link DataStreamWriter} object.<br />
 * <br />
 * An example of how to use this class can be found in
 * {@link org.LiveGraph.demoDataSource.LiveGraphDemo}.<br />
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: DataStreamWriterFactory.java</p> 
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p>
 * 
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class DataStreamWriterFactory {

/**
 * Data file extension used if no other specified.
 */
public static final String defaultFileExtension = "dat";


/**
 * Creates a new {@link DataStreamWriter} object for a unique file located in the
 * specified directory. The file will be named on the basis of the specified name and the
 * current date and time.<br />
 * The file name will be constructed according to the following pattern:<br />
 * {@code dataFileName.YY.MM.DD-HH.MM.SS.}{@value #defaultFileExtension}
 * 
 * @param dataFileDir The local folder in thich the file will be located.
 * @param dataFileName The basis for the file name.
 * @return A {@code DataStreamWriter} for the file with the name described above.
 */
public static DataStreamWriter createDataWriter(String dataFileDir, String dataFileName) {
	return createDataWriter(dataFileDir, dataFileName, defaultFileExtension);
}

/**
 * Creates a new {@link DataStreamWriter} object for a unique file located in the
 * specified directory. The file will be named on the basis of the specified name, the
 * specified extension and the current date and time.<br />
 * The file name will be constructed according to the following pattern:<br />
 * {@code dataFileName.YY.MM.DD-HH.MM.SS.dataFileExt}
 * 
 * @param dataFileDir The local folder in thich the file will be located.
 * @param dataFileName The basis for the file name.
 * @param dataFileExt The extension for the file name.
 * @return A {@code DataStreamWriter} for the file with the name described above.
 */
public static DataStreamWriter createDataWriter(String dataFileDir, String dataFileName, String dataFileExt) {
	File file = getUniqueDataFile(dataFileDir, dataFileName, dataFileExt);
	final int MAX_RETRY = 5;
	for (int retry = 0; retry < MAX_RETRY; retry++) {
		
		try {
			FileOutputStream outs = new FileOutputStream(file);
			DataStreamWriter writer = new DataStreamWriter(outs);
			return writer;
		} catch (FileNotFoundException e) {
			Thread.yield();
			try {Thread.sleep((int) (Math.random() * 1000)); }
			catch (InterruptedException e2) {}
			Thread.yield();
		}				
	}
	throw new Error("Cannot create data writer. Something weird is happening concurrently."); 
}


/**
 * Constructs a unique file name for a file located in the specified directory. The file will
 * be named on the basis of the specified name, the specified extension and the current date
 * and time.<br />
 * The file name will be constructed according to the following pattern:<br />
 * {@code dataFileName.YY.MM.DD-HH.MM.SS.dataFileExt} <br />
 * If that file already exists, the file name will be adjusted to <br />
 * {@code dataFileName.YY.MM.DD-HH.MM.SS(1).dataFileExt}, <br />
 * {@code dataFileName.YY.MM.DD-HH.MM.SS(2).dataFileExt}, <br />
 * and so on until the file does not already exist.
 * 
 * @param dataFileDir The local folder in thich the file will be located.
 * @param dataFileName The basis for the file name.
 * @param dataFileExt The extension for the file name.
 * @return A file object describing a file with the name as specified above.
 */
public static File getUniqueDataFile(String dataFileDir, String dataFileName, String dataFileExt) {
	
	if (null == dataFileDir)
		throw new NullPointerException("Data file directory cannot be null");
	
	if (null == dataFileName)
		throw new NullPointerException("Data file name basis cannot be null");
	
	if (null == dataFileExt)
		throw new NullPointerException("Data file extension cannot be null");
	
	if (dataFileDir.trim().length() > 0 &&  !dataFileDir.endsWith("/") && !dataFileDir.endsWith("\\"))
		dataFileDir = dataFileDir + "/";
	
	while (dataFileDir.length() > 1 && dataFileDir.startsWith("."))
		dataFileDir = dataFileDir.substring(1);
	
	Calendar date = new GregorianCalendar();
	String fName = String.format("%s.%2$ty.%2$tm.%2$td-%2$tH.%2$tM.%2$tS.%3$s", dataFileName, date, dataFileExt);
	File dataFile = new File(dataFileDir + fName);
	if (dataFile.exists()) {
		int k = 1;
		while(dataFile.exists()) {
			fName = String.format("%1$s.%2$ty.%2$tm.%2$td-%2$tH.%2$tM.%2$tS(%4$d).%3$s",
								  dataFileName, date, dataFileExt, k);
			dataFile = new File(dataFileDir + fName);
			k++;
		}
	}
	
	return dataFile;
}

} // public class DataStreamWriterFactory
